# define USE_RINTERNALS


#include <Rmath.h>
#include <math.h>
#include <R.h> 

#include <Rinternals.h>


double huge = 1.0e300;
static int *data1, *data2;



void fitch_free(){
    free(data1);
    free(data2);
}



void fitch_init(int *data, int *m, int *n)
{
    int i;
    data1 = (int *) calloc(*n, sizeof(int));
    data2 = (int *) calloc(*n, sizeof(int));
    for(i=0; i<*m; i++)data1[i] = data[i];
}



SEXP getData(SEXP n){
    int i, m=INTEGER(n)[0];  
    SEXP DAT, DAT2, RESULT;
    PROTECT(RESULT = allocVector(VECSXP, 2L));
    PROTECT(DAT = allocVector(INTSXP, m));
    PROTECT(DAT2 = allocVector(INTSXP, m)); 
    for(i=0; i<m; i++) INTEGER(DAT)[i] = data1[i];
    for(i=0; i<m; i++) INTEGER(DAT2)[i] = data2[i];
    SET_VECTOR_ELT(RESULT, 0, DAT);
    SET_VECTOR_ELT(RESULT, 1, DAT2);
    UNPROTECT(3);
    return(RESULT); 
}



int bitcount(int x){ 
    int count;
    for (count=0; x != 0; x>>=1)
       if ( x & 01)
           count++;
    return count;
}


void bitCount(int *x, int *count){
    count[0]=bitcount(x[0]);
} 




void fitch43(int *dat1, int *dat2, int *nr, int *pars, double *weight, double *w){
    int k, tmp;
    for(k = 0; k < (*nr); k++){
        tmp = dat1[k] & dat2[k];
        if(!tmp){
            tmp = dat1[k] | dat2[k];
            (pars[k])++;
            (*w)+=weight[k];
        }
        dat1[k] = tmp;
    } 
}


void fitch44(int *res, int *dat1, int *dat2, int *nr, int *pars, double *weight, double *w){
    int k, tmp;
    for(k = 0; k < (*nr); k++){
        tmp = dat1[k] & dat2[k];
        if(!tmp){
            tmp = dat1[k] | dat2[k];
            (pars[k])++;
            (*w)+=weight[k];
        }
        res[k] = tmp;
    } 
}


 
SEXP FITCHTRIP(SEXP DAT3, SEXP nrx, SEXP edge, SEXP weight){ 
    R_len_t i, m = length(edge);  
    int nr=INTEGER(nrx)[0], k, tmp, ei;  
    int *dat3=INTEGER(DAT3);
    double *pvtmp;  
    SEXP pvec;
    PROTECT(pvec = allocVector(REALSXP, m));
    pvtmp = REAL(pvec);
    for(i=0; i<m; i++) pvtmp[i] = 0.0;
    for(i=0; i<m; i++){
        ei = INTEGER(edge)[i] - 1L;
        for(k = 0; k < nr; k++){
            tmp = data1[k + ei*nr] & data2[k + ei*nr];
            if(!tmp){
                tmp = data1[k + ei*nr] | data2[k + ei*nr];
                REAL(pvec)[i]+=REAL(weight)[k];
            }
            tmp = tmp & dat3[k];
            if(!tmp){
               REAL(pvec)[i]+=REAL(weight)[k];
            }
        }
    }
    UNPROTECT(1);
    return(pvec); 
}


//  .C version of FITCHTRIP  index statt dat3
void FITCHTRIP2(int *dat3, int *nr, int *edge, int *weight, int *m, double *pvec, double *p0){ 
    int i, k, tmp, ei;  
    for(i=0; i<*m; i++){
        ei = INTEGER(edge)[i] - 1L;
        for(k = 0; k < *nr; k++){
            tmp = data1[k + (ei* *nr)] & data2[k + (ei* *nr)];
            if(!tmp){
                tmp = data1[k + (ei* *nr)] | data2[k + (ei* *nr)];
                pvec[i]+=weight[k];
                if(pvec[i]>*p0)break;
            }
            tmp = tmp & dat3[k];
            if(!tmp){
                pvec[i]+=weight[k];
                if(pvec[i]>*p0)break;
            }
        }
        if(pvec[i]<*p0) p0[0]=pvec[i]; //nicht so wichtig
    }
}




void fitch6(int *dat, int *nr, int *pars, int *node, int *edge, int *nl, double *weight, double *pvec, double *pscore) 
{   
    int i, ni, k;
    ni = 0;
    for (i=0; i< *nl; i++) {
	if (ni == node[i]){
	         pvec[ni-1L] += pvec[edge[i]-1L];
	         fitch43(&dat[(ni-1L) * (*nr)], &dat[(edge[i]-1L) * (*nr)], nr, pars, weight, &pvec[(ni-1L)]); 
        }                  
        else {
            ni = node[i];   
            pvec[(ni-1L)] += pvec[(edge[i]-1L)];         
            for(k = 0; k < (*nr); k++) dat[(ni-1L)*(*nr) + k] = dat[(edge[i]-1L)*(*nr) + k];                     
        }
    }
    pscore[0]=pvec[ni-1];
}



SEXP FITCH(SEXP dat, SEXP nrx, SEXP node, SEXP edge, SEXP l, SEXP weight, SEXP mx, SEXP q){   
    int *data, *nr=INTEGER(nrx), m=INTEGER(mx)[0], i, n=INTEGER(q)[0];   
    double *pvtmp;  
    SEXP DAT, pars, pvec, pscore, RESULT;
    PROTECT(RESULT = allocVector(VECSXP, 4L));
    PROTECT(pars = allocVector(INTSXP, *nr));
    PROTECT(pscore = allocVector(REALSXP, 1L));
    PROTECT(DAT = allocMatrix(INTSXP, nr[0], m));
    PROTECT(pvec = allocVector(REALSXP, m));
    pvtmp = REAL(pvec);
    data = INTEGER(DAT);
    for(i=0; i<m; i++) pvtmp[i] = 0.0;
    for(i=0; i<*nr; i++) INTEGER(pars)[i] = 0L;
    REAL(pscore)[0]=0.0;
    for(i=0; i<(*nr * n); i++)data[i] = INTEGER(dat)[i];
    
    fitch6(data, nr, INTEGER(pars), INTEGER(node), INTEGER(edge), INTEGER(l), REAL(weight), pvtmp, REAL(pscore));
    
    SET_VECTOR_ELT(RESULT, 0, pscore);
    SET_VECTOR_ELT(RESULT, 1, pars);
    SET_VECTOR_ELT(RESULT, 2, DAT);
    SET_VECTOR_ELT(RESULT, 3, pvec);
    UNPROTECT(5);
    return(RESULT); 
}


// requires binary tree
void fitch8(int *dat, int *nr, int *pars, int *node, int *edge, int *nl, double *weight, double *pvec, double *pscore) 
{   
    int i, ni=0L, ri, le;
    i=0L;
    while(i<(*nl - 1L)){
        ni = node[i] - 1L; 
        le = edge[i] - 1L;
        ri = edge[i+1] - 1L; 
        pvec[ni] = pvec[le] + pvec[ri];
	fitch44(&dat[ni * (*nr)], &dat[le * (*nr)], &dat[ri * (*nr)], nr, pars, weight, &pvec[ni]); 
        i++;
        i++;                  
    }
    if(i == (*nl-1L)){
        le = edge[i] - 1L;
        pvec[ni] += pvec[le]; 
        fitch43(&dat[ni * (*nr)], &dat[le * (*nr)], nr, pars, weight, &pvec[ni]); 
    } 
    pscore[0]=pvec[ni];
}



/*
ACCTRAN
*/
void fitchT(int *dat1, int *dat2, int *nr, double *pars, double *weight, double *w){
    int k;
    int tmp;
    for(k = 0; k < (*nr); k++){
        tmp = dat1[k] & dat2[k];
        if(tmp > 0L){
             dat1[k] = tmp;
             }
    } 
}


void fitchT3(int *dat1, int *dat2, int *nr, double *pars, double *weight, double *w){
    int k;
    int tmp;
    for(k = 0; k < (*nr); k++){
       tmp = dat1[k] & dat2[k];
       if(tmp==0L) {
             (*w)+=weight[k];
             pars[k] += 1;
             }
       if(tmp >0){
           if(tmp < dat2[k]){ 
              (*w)+= .5*weight[k];
              pars[k] += .5;
           }
       }

    } 
}


// return lower and upper bound for the number of changes 
// upper bound very conservative 
void countMPR(double *res, int *dat1, int *dat2, int *nr, double *weight, int *external){
    int k;
    int tmp;
    for(k = 0; k < (*nr); k++){
        tmp = dat1[k] & dat2[k];

        if(tmp==0){
            res[0] += weight[k];
            res[1] += weight[k];
        }
        else{ 
            if( external[0]==0L){ 
                 if( bitcount(dat1[k] | dat2[k])>1L ) res[1] += weight[k]; // dat1[k] != dat2[k]
            }   
            else{ 
                 if( tmp  < dat2[k] ) res[1] += weight[k];
            }
        }
    } 
}


void ACCTRAN2(int *dat, int *nr, double *pars, int *node, int *edge, int *nl, double *weight, double *pvec, int *nTips) 
{   
    int i;
    for (i=0; i< *nl; i++) {       
        if(edge[i]>nTips[0]) fitchT(&dat[(edge[i]-1L) * (*nr)], &dat[(node[i]-1) * (*nr)], nr, pars, weight, &pvec[i]); 
        }
}


void ACCTRAN3(int *dat, int *nr, double *pars, int *node, int *edge, int *nl, double *weight, double *pvec, int *nTips) 
{   
    int i;
    for (i=0; i< *nr; i++)pars[i]=0.0;
    for(i=0; i< *nl; i++)pvec[i] = 0.0;
    for (i=0; i< *nl; i++) {               
        fitchT3(&dat[(edge[i]-1L) * (*nr)], &dat[(node[i]-1) * (*nr)], nr, pars, weight, &pvec[i]); 
    }            
}


void fitchN(int *dat1, int *dat2, int *nr){
    int k;
    int tmp;
    for(k = 0; k < (*nr); k++){
        tmp = dat1[k] & dat2[k];
        if(tmp) dat1[k] = tmp;
        else dat1[k] = dat1[k] | dat2[k];
        } 
}

// raus
void fitchN2(int *res, int *dat, int *node, int *edge, int *nr, int *nl) { 
    int i;
    for (i=0; i< *nl; i++) {
        fitchN(&res[(node[i]-1L) * (*nr)], &dat[(edge[i]-1L) * (*nr)], nr);              
    }
}


void fitchTriplet(int *res, int *dat1, int *dat2, int *dat3, int *nr) 
{   
    int k; // ni,
//    ni = 0;
    
    int *v1, *v2, *v3;
    v1 = (int *) R_alloc(*nr, sizeof(int));    
    v2 = (int *) R_alloc(*nr, sizeof(int));
    v3 = (int *) R_alloc(*nr, sizeof(int));

    for(k = 0; k < (*nr); k++) v1[k] = dat1[k];
    fitchN(v1, dat2, nr);
    fitchN(v1, dat3, nr);

    for(k = 0; k < (*nr); k++) v2[k] = dat1[k];
    fitchN(v2, dat3, nr);
    fitchN(v2, dat2, nr);

    for(k = 0; k < (*nr); k++) v3[k] = dat2[k];
    fitchN(v3, dat3, nr);
    fitchN(v3, dat1, nr);

    for(k = 0; k < (*nr); k++)res[k] = v1[k] & v2[k]; // &v3[k];  
    for(k = 0; k < (*nr); k++)res[k] = res[k] & v3[k];  
}


void fitchNACC2(int *root, int *dat, int *nr, double *pars, int *result, double *weight, double *pars1){
    int k;
    int tmp;
    for(k = 0; k < (*nr); k++){
//       result[k] = 0L;
       tmp = root[k] & dat[k];
       if(tmp==0L) {
             pars[0] += weight[k];
             pars1[k] += weight[k];
             }
       if(tmp >0){
           if(tmp < root[k]){ 
              pars[0] += .5*weight[k];
              pars1[k] += .5*weight[k];
              result[k] += 1L;
           }
       }
    }        
}


void fitchTripletACC4(int *root, int *dat1, int *dat2, int *dat3, int *nr, double *p1, double *p2, double *p3, double *weight, double *pars1, int *v1) 
{   
    int k;
       
    int tmp, a, b, c, t1, t2, t3;
    double d, f;
    for(k = 0; k < (*nr); k++){
        tmp = root[k];
        a = dat1[k] & dat2[k]; 
        b = dat1[k] & dat3[k];
        c = dat2[k] & dat3[k];
        if((a+b+c) == 0L){
           d = (2.0/3.0) * weight[k];
           p1[0] += d; 
           p2[0] += d;
           p3[0] += d;        
           pars1[k] += 2*weight[k]; 
           v1[k] = 2L; 
        }
        else{  
            f = 0.0;
            d = weight[k];
            t1 = 0.0;
            t2 = 0.0;
            t3 = 0.0;
            if( (dat1[k] & tmp)<tmp){ 
                t1 = d; 
                f+=1.0;
            }
            if( (dat2[k] & tmp)<tmp){ 
                t2 = d; 
                f+=1.0;
            }
            if( (dat3[k] & tmp)<tmp){ 
                t3 = d; 
                f+=1.0;
            }
            if(f>0.0){   
                pars1[k] += weight[k]; 
                p1[0] += t1/f; 
                p2[0] += t2/f;
                p3[0] += t3/f;
                v1[k] += 1L;
            }
        }
    }
}


// raus
SEXP FITCH123(SEXP dat, SEXP nrx, SEXP node, SEXP edge, SEXP l, SEXP weight, SEXP mx, SEXP q, SEXP ps){   
    int *data, *nr=INTEGER(nrx), m=INTEGER(mx)[0], i, n=INTEGER(q)[0];   //=INTEGER(dat)
    double *pvtmp;  
    SEXP pars, pscore; 
    PROTECT(pars = allocVector(INTSXP, *nr));
    PROTECT(pscore = allocVector(REALSXP, 1L));
    pvtmp = (double *) R_alloc(m, sizeof(double)); 
    data = (int *) R_alloc(nr[0]*m, sizeof(int)); 
    for(i=0; i<m; i++) pvtmp[i] = 0.0;
    for(i=0; i<*nr; i++) INTEGER(pars)[i] = 0L;
    REAL(pscore)[0]=0.0;
    for(i=0; i<(*nr * n); i++)data[i] = INTEGER(dat)[i];
    for(i=(*nr * n); i<(*nr * m); i++)data[i] = 0L;
    fitch6(data, nr, INTEGER(pars), INTEGER(node), INTEGER(edge), INTEGER(l), REAL(weight), pvtmp, REAL(pscore));
    UNPROTECT(2);
    if(INTEGER(ps)[0]==1)return(pscore);
    else return(pars); 
}



SEXP FITCH345(SEXP nrx, SEXP node, SEXP edge, SEXP l, SEXP weight, SEXP mx, SEXP ps){   
    int *nr=INTEGER(nrx), m=INTEGER(mx)[0], i;  
    double *pvtmp;  
    SEXP pars, pscore; 
    PROTECT(pars = allocVector(INTSXP, *nr));
    PROTECT(pscore = allocVector(REALSXP, 1L));
    pvtmp = (double *) R_alloc(m, sizeof(double)); 
    for(i=0; i<m; i++) pvtmp[i] = 0.0;
    for(i=0; i<*nr; i++) INTEGER(pars)[i] = 0L;
    REAL(pscore)[0]=0.0;
    fitch8(data1, nr, INTEGER(pars), INTEGER(node), INTEGER(edge), INTEGER(l), REAL(weight), pvtmp, REAL(pscore));
    UNPROTECT(2);
    if(INTEGER(ps)[0]==1)return(pscore);
    else return(pars); 
}



void FN3(int *dat, int *res, int *nr, int *pars, int *node, int *edge, int *nl, int *pc, double *weight, double *tmpvec, double *pvec) { 
    int i=0L, ni, le, ri;
    while(i< *nl) {
        ni = node[i] - 1L;
        le = edge[i] - 1L;
        ri = edge[i+1L] - 1L;
        if(pc[i+1L]==0L){
	    pvec[ni] = tmpvec[le] + tmpvec[ri];
	    fitch44(&res[ni * (*nr)], &dat[(edge[i]-1L) * (*nr)], &dat[ri * (*nr)], nr, pars, weight, &pvec[ni]);              
        }    
        else{ 
            pvec[ni] = tmpvec[le] + pvec[ri];
	    fitch44(&res[ni * (*nr)], &dat[le * (*nr)], &res[ri * (*nr)], nr, pars, weight, &pvec[ni]);   
        }
        i++;
        i++;
    }
}



void FN2(int *dat, int *res, int *nr, int *pars, int *node, int *edge, int *nl, int *pc, double *weight, double *tmpvec, double *pvec) { 
    int i, ni, k;
    ni = 0L;
    for (i=0; i< *nl; i++) {
        if (ni == node[i]){
              if(pc[i]==0L){
	              pvec[(ni-1L)] += tmpvec[(edge[i]-1L)];
	              fitch43(&res[(ni-1L) * (*nr)], &dat[(edge[i]-1L) * (*nr)], nr, pars, weight, &pvec[(ni-1L)]);              
              }    
              else{ 
                      pvec[(ni-1L)] += pvec[(edge[i]-1L)];
	              fitch43(&res[(ni-1L) * (*nr)], &res[(edge[i]-1L) * (*nr)], nr, pars, weight, &pvec[(ni-1L)]);   
              }
        }                      
        else {
            ni = node[i];
            pvec[(ni-1L)] += tmpvec[(edge[i]-1L)];
            for(k = 0; k < (*nr); k++) res[(ni-1L)*(*nr) + k] = dat[(edge[i]-1L)*(*nr) + k];                 
        }
    }
}



// raus
SEXP FNALL(SEXP dat, SEXP nrx, SEXP node, SEXP edge, SEXP node2, SEXP edge2, SEXP l, SEXP weight, SEXP mx, SEXP my, SEXP q, SEXP pc){   
    int *data=INTEGER(dat), *nr=INTEGER(nrx), m=INTEGER(mx)[0], i, n=INTEGER(q)[0], *pars2, *dat2;  
    double *pvtmp, *pvtmp2;  
    SEXP DAT, DAT2, pars, pvec, pvec2, pscore, RESULT;
  
    pars2 = (int *) R_alloc(*nr, sizeof(int));    

    PROTECT(RESULT = allocVector(VECSXP, 6L));
    PROTECT(pars = allocVector(INTSXP, *nr));
    PROTECT(pscore = allocVector(REALSXP, 1L));
    PROTECT(DAT = allocMatrix(INTSXP, nr[0], m));
    PROTECT(DAT2 = allocMatrix(INTSXP, nr[0], m));
    PROTECT(pvec = allocVector(REALSXP, m));
    PROTECT(pvec2 = allocVector(REALSXP, m)); 
    for(i=0; i<*nr; i++) INTEGER(pars)[i] = 0L;
    pvtmp = REAL(pvec);
    for(i=0; i<m; i++) pvtmp[i] = 0.0;
    pvtmp2 = REAL(pvec2);
    for(i=0; i<m; i++) pvtmp2[i] = 0.0;
    REAL(pscore)[0]=0.0;
    for(i=0; i<(*nr * n); i++)INTEGER(DAT)[i] = data[i];
    dat2 = INTEGER(DAT2); // without, wired things happen
    for(i=0; i<(*nr * m); i++)dat2[i] = 0L;
    fitch6(INTEGER(DAT), nr, INTEGER(pars), INTEGER(node), INTEGER(edge), INTEGER(l), REAL(weight), pvtmp, REAL(pscore));   

    FN2(INTEGER(DAT), dat2, nr, pars2, INTEGER(node2), INTEGER(edge2), INTEGER(my), INTEGER(pc), REAL(weight), pvtmp, pvtmp2);

    SET_VECTOR_ELT(RESULT, 0, pscore);
    SET_VECTOR_ELT(RESULT, 1, pars);
    SET_VECTOR_ELT(RESULT, 2, DAT);
    SET_VECTOR_ELT(RESULT, 3, DAT2);
    SET_VECTOR_ELT(RESULT, 4, pvec);
    SET_VECTOR_ELT(RESULT, 5, pvec2);
    UNPROTECT(7);
    return(RESULT); 
}



SEXP FNALL3(SEXP nrx, SEXP node, SEXP edge, SEXP node2, SEXP edge2, SEXP l, SEXP weight, SEXP mx, SEXP my, SEXP q, SEXP pc){   
    int *nr=INTEGER(nrx), m=INTEGER(mx)[0], i, *pars2;  
    double *pvtmp, *pvtmp2;  
    SEXP pars, pvec, pvec2, pscore, RESULT;
  
    pars2 = (int *) R_alloc(*nr, sizeof(int));    

    PROTECT(RESULT = allocVector(VECSXP, 4L));
    PROTECT(pars = allocVector(INTSXP, *nr));
    PROTECT(pscore = allocVector(REALSXP, 1L));
    PROTECT(pvec = allocVector(REALSXP, m));
    PROTECT(pvec2 = allocVector(REALSXP, m)); 
    for(i=0; i<*nr; i++) INTEGER(pars)[i] = 0L;
    pvtmp = REAL(pvec);
    for(i=0; i<m; i++) pvtmp[i] = 0.0;
    pvtmp2 = REAL(pvec2);
    for(i=0; i<m; i++) pvtmp2[i] = 0.0;
    REAL(pscore)[0]=0.0;
    fitch8(data1, nr, INTEGER(pars), INTEGER(node), INTEGER(edge), INTEGER(l), REAL(weight), pvtmp, REAL(pscore));   
    FN3(data1, data2, nr, pars2, INTEGER(node2), INTEGER(edge2), INTEGER(my), INTEGER(pc), REAL(weight), pvtmp, pvtmp2);

    SET_VECTOR_ELT(RESULT, 0, pscore);
    SET_VECTOR_ELT(RESULT, 1, pars);
    SET_VECTOR_ELT(RESULT, 2, pvec);
    SET_VECTOR_ELT(RESULT, 3, pvec2);
    UNPROTECT(5);
    return(RESULT); 
}


// inside optNNI Ziel 3* schneller  , double best
void fitchquartet(int *dat1, int *dat2, int *dat3, int *dat4, int *nr, double *weight, double *pars){   
    int k, tmp1, tmp2;  
    pars[0] = 0.0; 
    for(k = 0; k < *nr; k++){
        tmp1 = dat1[k] & dat2[k];
        tmp2 = dat3[k] & dat4[k];  
        if(!tmp1){
            tmp1 = dat1[k] | dat2[k];
            pars[0]+=weight[k];
        }
        if(!tmp2){
            tmp2 = dat3[k] | dat4[k];
            pars[0]+=weight[k];
        }
        tmp1 = tmp1 & tmp2;
        if(!tmp1){
            pars[0]+=weight[k];
        }
    }
}


void fitchQuartet(int *index, int *n, int *nr, double *psc1, double *psc2, double *weight, double *res){
    int i, e1, e2, e3, e4;
    for(i=0; i<*n; i++){ 
        e1=index[(i* 6L)] - 1L;
        e2=index[1L + (i* 6L)] - 1L;
        e3=index[2L + (i* 6L)] - 1L;
        e4=index[3L + (i* 6L)] - 1L;

        if(index[5L + (i* 6L)] == 1){
            fitchquartet(&data2[e1 * (*nr)], &data1[e2 * (*nr)], &data1[e3 * (*nr)], &data1[e4 * (*nr)], nr, weight, &res[i]);
            res[i] += psc2[e1] + psc1[e2] + psc1[e3] + psc1[e4]; // stimmt
        } 
        else{
            fitchquartet(&data1[e1 * (*nr)], &data1[e2 * (*nr)], &data1[e3 * (*nr)], &data1[e4 * (*nr)], nr, weight, &res[i]);
            res[i] += psc1[e1] + psc1[e2] + psc1[e3] + psc1[e4]; 
        }
    }
}




